@@ -55,25 +55,33 @@ typedef struct StackFrame {
 
 GC_INNER ptr_t GC_FindTopOfStack(unsigned long stack_start)
 {
-  StackFrame *frame;
+  StackFrame *frame = (StackFrame *)stack_start;
 
-# ifdef POWERPC
-    if (stack_start == 0) {
+  if (stack_start == 0) {
+#   ifdef POWERPC
 #     if CPP_WORDSZ == 32
         __asm__ __volatile__ ("lwz %0,0(r1)" : "=r" (frame));
 #     else
         __asm__ __volatile__ ("ld %0,0(r1)" : "=r" (frame));
 #     endif
-    } else
-# else
-    GC_ASSERT(stack_start != 0); /* not implemented */
-# endif /* !POWERPC */
-  /* else */ {
-    frame = (StackFrame *)stack_start;
+#   elif defined(ARM32)
+        volatile ptr_t sp_reg;
+        __asm__ __volatile__ ("mov %0, r7\n" : "=r" (sp_reg));
+        frame = (StackFrame *)sp_reg;
+#   elif defined(AARCH64)
+        volatile ptr_t sp_reg;
+        __asm__ __volatile__ ("mov %0, x29\n" : "=r" (sp_reg));
+        frame = (StackFrame *)sp_reg;
+#   else
+#     if defined(CPPCHECK)
+        GC_noop1((word)&frame);
+#     endif
+      ABORT("GC_FindTopOfStack(0) is not implemented");
+#   endif
   }
 
 # ifdef DEBUG_THREADS_EXTRA
-    GC_log_printf("FindTopOfStack start at sp = %p\n", (void *)frame);
+    GC_log_printf("FindTopOfStack start at sp= %p\n", (void *)frame);
 # endif
   while (frame->savedSP != 0) {
     /* if there are no more stack frames, stop */
@@ -87,7 +95,7 @@ GC_INNER ptr_t GC_FindTopOfStack(unsigned long stack_start)
       break; /* if the next LR is bogus, stop */
   }
 # ifdef DEBUG_THREADS_EXTRA
-    GC_log_printf("FindTopOfStack finish at sp = %p\n", (void *)frame);
+    GC_log_printf("FindTopOfStack finish at sp= %p\n", (void *)frame);
 # endif
   return (ptr_t)frame;
 }
@@ -185,25 +193,27 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
             ABORT("unified_state flavor should be ARM_THREAD_STATE32");
           }
 #       endif
-        state = unified_state.ts_32;
+        state = unified_state;
       } else
 #   endif
     /* else */ {
       mach_msg_type_number_t thread_state_count = GC_MACH_THREAD_STATE_COUNT;
 
-      /* Get the thread state (registers, etc) */
-      kern_result = thread_get_state(thread, GC_MACH_THREAD_STATE,
-                                     (natural_t *)&state,
-                                     &thread_state_count);
+      /* Get the thread state (registers, etc.) */
+      do {
+        kern_result = thread_get_state(thread, GC_MACH_THREAD_STATE,
+                                       (natural_t *)&state,
+                                       &thread_state_count);
+      } while (kern_result == KERN_ABORTED);
     }
 #   ifdef DEBUG_THREADS
-      GC_log_printf("thread_get_state returns value = %d\n", kern_result);
+      GC_log_printf("thread_get_state returns %d\n", kern_result);
 #   endif
     if (kern_result != KERN_SUCCESS)
       ABORT("thread_get_state failed");
 
 #   if defined(I386)
-      lo = (void *)state.THREAD_FLD(esp);
+      lo = (ptr_t)state.THREAD_FLD(esp);
 #     ifndef DARWIN_DONT_PARSE_STACK
         *phi = GC_FindTopOfStack(state.THREAD_FLD(esp));
 #     endif
@@ -216,7 +226,7 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
       GC_push_one(state.THREAD_FLD(ebp));
 
 #   elif defined(X86_64)
-      lo = (void *)state.THREAD_FLD(rsp);
+      lo = (ptr_t)state.THREAD_FLD(rsp);
 #     ifndef DARWIN_DONT_PARSE_STACK
         *phi = GC_FindTopOfStack(state.THREAD_FLD(rsp));
 #     endif
@@ -227,7 +237,7 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
       GC_push_one(state.THREAD_FLD(rdi));
       GC_push_one(state.THREAD_FLD(rsi));
       GC_push_one(state.THREAD_FLD(rbp));
-      /* GC_push_one(state.THREAD_FLD(rsp)); */
+      /* rsp is skipped.        */
       GC_push_one(state.THREAD_FLD(r8));
       GC_push_one(state.THREAD_FLD(r9));
       GC_push_one(state.THREAD_FLD(r10));
@@ -238,11 +248,12 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
       GC_push_one(state.THREAD_FLD(r15));
 
 #   elif defined(POWERPC)
-      lo = (void *)(state.THREAD_FLD(r1) - PPC_RED_ZONE_SIZE);
+      lo = (ptr_t)(state.THREAD_FLD(r1) - PPC_RED_ZONE_SIZE);
 #     ifndef DARWIN_DONT_PARSE_STACK
         *phi = GC_FindTopOfStack(state.THREAD_FLD(r1));
 #     endif
       GC_push_one(state.THREAD_FLD(r0));
+      /* r1 is skipped. */
       GC_push_one(state.THREAD_FLD(r2));
       GC_push_one(state.THREAD_FLD(r3));
       GC_push_one(state.THREAD_FLD(r4));
@@ -275,24 +286,25 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
       GC_push_one(state.THREAD_FLD(r31));
 
 #   elif defined(ARM32)
-      lo = (void *)state.THREAD_FLD(sp);
+      lo = (ptr_t)state.THREAD_FLD(sp);
 #     ifndef DARWIN_DONT_PARSE_STACK
-        *phi = GC_FindTopOfStack(state.THREAD_FLD(sp));
+        *phi = GC_FindTopOfStack(state.THREAD_FLD(r[7])); /* fp */
 #     endif
       {
         int j;
-        for (j = 0; j <= 12; j++) {
+        for (j = 0; j < 7; j++)
+          GC_push_one(state.THREAD_FLD(r[j]));
+        j++; /* "r7" is skipped (iOS uses it as a frame pointer) */
+        for (; j <= 12; j++)
           GC_push_one(state.THREAD_FLD(r[j]));
-        }
       }
-      /* "pc" and "sp" are skipped */
+      /* "cpsr", "pc" and "sp" are skipped */
       GC_push_one(state.THREAD_FLD(lr));
-      GC_push_one(state.THREAD_FLD(cpsr));
 
 #   elif defined(AARCH64)
-      lo = (void *)state.THREAD_FLD(sp);
+      lo = (ptr_t)state.THREAD_FLD(sp);
 #     ifndef DARWIN_DONT_PARSE_STACK
-        *phi = GC_FindTopOfStack(state.THREAD_FLD(sp));
+        *phi = GC_FindTopOfStack(state.THREAD_FLD(fp));
 #     endif
       {
         int j;
@@ -300,14 +312,13 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
           GC_push_one(state.THREAD_FLD(x[j]));
         }
       }
-      /* "cpsr", "pc" and "sp" are skipped */
-      GC_push_one(state.THREAD_FLD(fp));
+      /* "cpsr", "fp", "pc" and "sp" are skipped */
       GC_push_one(state.THREAD_FLD(lr));
 
 #   elif defined(CPPCHECK)
       lo = NULL;
 #   else
-#     error FIXME for non-x86 || ppc || arm architectures
+#     error FIXME for non-arm/ppc/x86 architectures
 #   endif
   } /* thread != my_thread */
 
@@ -330,7 +341,7 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
     *paltstack_lo = NULL;
   }
 # ifdef DEBUG_THREADS
-    GC_log_printf("Darwin: Stack for thread %p = [%p,%p)\n",
+    GC_log_printf("Darwin: Stack for thread %p is [%p,%p)\n",
                   (void *)(word)thread, (void *)lo, (void *)(*phi));
 # endif
   return lo;
@@ -436,7 +447,7 @@ GC_INNER void GC_push_all_stacks(void)
 
   struct GC_mach_thread {
     thread_act_t thread;
-    GC_bool already_suspended;
+    GC_bool suspended;
   };
 
   struct GC_mach_thread GC_mach_threads[GC_MAX_MACH_THREADS];
@@ -446,7 +457,8 @@ GC_INNER void GC_push_all_stacks(void)
 /* returns true if there's a thread in act_list that wasn't in old_list */
 STATIC GC_bool GC_suspend_thread_list(thread_act_array_t act_list, int count,
                                       thread_act_array_t old_list,
-                                      int old_count, mach_port_t my_thread)
+                                      int old_count, task_t my_task,
+                                      mach_port_t my_thread)
 {
   int i;
   int j = -1;
@@ -455,27 +467,22 @@ STATIC GC_bool GC_suspend_thread_list(thread_act_array_t act_list, int count,
   for (i = 0; i < count; i++) {
     thread_act_t thread = act_list[i];
     GC_bool found;
-    struct thread_basic_info info;
-    mach_msg_type_number_t outCount;
     kern_return_t kern_result;
 
     if (thread == my_thread
 #       ifdef MPROTECT_VDB
           || (GC_mach_handler_thread == thread && GC_use_mach_handler_thread)
+#       endif
+#       ifdef PARALLEL_MARK
+          || GC_is_mach_marker(thread) /* ignore the parallel markers */
 #       endif
         ) {
-      /* Don't add our and the handler threads. */
+      /* Do not add our one, parallel marker and the handler threads;   */
+      /* consider it as found (e.g., it was processed earlier).         */
+      mach_port_deallocate(my_task, thread);
       continue;
     }
-#   ifdef PARALLEL_MARK
-      if (GC_is_mach_marker(thread))
-        continue; /* ignore the parallel marker threads */
-#   endif
 
-#   ifdef DEBUG_THREADS
-      GC_log_printf("Attempting to suspend thread %p\n",
-                    (void *)(word)thread);
-#   endif
     /* find the current thread in the old list */
     found = FALSE;
     {
@@ -494,55 +501,44 @@ STATIC GC_bool GC_suspend_thread_list(thread_act_array_t act_list, int count,
             found = TRUE;
             break;
           }
-
-        if (!found) {
-          /* add it to the GC_mach_threads list */
-          if (GC_mach_threads_count == GC_MAX_MACH_THREADS)
-            ABORT("Too many threads");
-          GC_mach_threads[GC_mach_threads_count].thread = thread;
-          /* default is not suspended */
-          GC_mach_threads[GC_mach_threads_count].already_suspended = FALSE;
-          changed = TRUE;
-        }
       }
     }
 
-    outCount = THREAD_INFO_MAX;
-    kern_result = thread_info(thread, THREAD_BASIC_INFO,
-                              (thread_info_t)&info, &outCount);
-    if (kern_result != KERN_SUCCESS) {
-      /* The thread may have quit since the thread_threads() call we  */
-      /* mark already suspended so it's not dealt with anymore later. */
-      if (!found)
-        GC_mach_threads[GC_mach_threads_count++].already_suspended = TRUE;
-      continue;
-    }
-#   ifdef DEBUG_THREADS
-      GC_log_printf("Thread state for %p = %d\n", (void *)(word)thread,
-                    info.run_state);
-#   endif
-    if (info.suspend_count != 0) {
-      /* thread is already suspended. */
-      if (!found)
-        GC_mach_threads[GC_mach_threads_count++].already_suspended = TRUE;
+    if (found) {
+      /* It is already in the list, skip processing, release mach port. */
+      mach_port_deallocate(my_task, thread);
       continue;
     }
 
+    /* add it to the GC_mach_threads list */
+    if (GC_mach_threads_count == GC_MAX_MACH_THREADS)
+      ABORT("Too many threads");
+    GC_mach_threads[GC_mach_threads_count].thread = thread;
+    /* default is not suspended */
+    GC_mach_threads[GC_mach_threads_count].suspended = FALSE;
+    changed = TRUE;
+
 #   ifdef DEBUG_THREADS
       GC_log_printf("Suspending %p\n", (void *)(word)thread);
 #   endif
-    kern_result = thread_suspend(thread);
+    /* Unconditionally suspend the thread.  It will do no     */
+    /* harm if it is already suspended by the client logic.   */
+    GC_acquire_dirty_lock();
+    do {
+      kern_result = thread_suspend(thread);
+    } while (kern_result == KERN_ABORTED);
+    GC_release_dirty_lock();
     if (kern_result != KERN_SUCCESS) {
       /* The thread may have quit since the thread_threads() call we  */
       /* mark already suspended so it's not dealt with anymore later. */
-      if (!found)
-        GC_mach_threads[GC_mach_threads_count++].already_suspended = TRUE;
-      continue;
+      GC_mach_threads[GC_mach_threads_count].suspended = FALSE;
+    } else {
+      /* Mark the thread as suspended and require resume.     */
+      GC_mach_threads[GC_mach_threads_count].suspended = TRUE;
+      if (GC_on_thread_event)
+        GC_on_thread_event(GC_EVENT_THREAD_SUSPENDED, (void *)(word)thread);
     }
-    if (!found)
-      GC_mach_threads_count++;
-    if (GC_on_thread_event)
-      GC_on_thread_event(GC_EVENT_THREAD_SUSPENDED, (void *)(word)thread);
+    GC_mach_threads_count++;
   }
   return changed;
 }
@@ -574,7 +570,6 @@ GC_INNER void GC_stop_world(void)
 
   if (GC_query_task_threads) {
 #   ifndef GC_NO_THREADS_DISCOVERY
-      unsigned i;
       GC_bool changed;
       thread_act_array_t act_list, prev_list;
       mach_msg_type_number_t listcount, prevcount;
@@ -598,12 +593,14 @@ GC_INNER void GC_stop_world(void)
 
         if (kern_result == KERN_SUCCESS) {
           changed = GC_suspend_thread_list(act_list, listcount, prev_list,
-                                           prevcount, my_thread);
+                                           prevcount, my_task, my_thread);
 
           if (prev_list != NULL) {
-            for (i = 0; i < prevcount; i++)
-              mach_port_deallocate(my_task, prev_list[i]);
-
+            /* Thread ports are not deallocated by list, unused ports   */
+            /* deallocated in GC_suspend_thread_list, used - kept in    */
+            /* GC_mach_threads till GC_start_world as otherwise thread  */
+            /* object change can occur and GC_start_world will not      */
+            /* find the thread to resume which will cause app to hang.  */
             vm_deallocate(my_task, (vm_address_t)prev_list,
                           sizeof(thread_t) * prevcount);
           }
@@ -615,8 +612,7 @@ GC_INNER void GC_stop_world(void)
       } while (changed);
 
       GC_ASSERT(prev_list != 0);
-      for (i = 0; i < prevcount; i++)
-        mach_port_deallocate(my_task, prev_list[i]);
+      /* The thread ports are not deallocated by list, see above.       */
       vm_deallocate(my_task, (vm_address_t)act_list,
                     sizeof(thread_t) * listcount);
 #   endif /* !GC_NO_THREADS_DISCOVERY */
@@ -630,8 +626,11 @@ GC_INNER void GC_stop_world(void)
       for (p = GC_threads[i]; p != NULL; p = p->next) {
         if ((p->flags & FINISHED) == 0 && !p->thread_blocked &&
              p->stop_info.mach_thread != my_thread) {
-
-          kern_result = thread_suspend(p->stop_info.mach_thread);
+          GC_acquire_dirty_lock();
+          do {
+            kern_result = thread_suspend(p->stop_info.mach_thread);
+          } while (kern_result == KERN_ABORTED);
+          GC_release_dirty_lock();
           if (kern_result != KERN_SUCCESS)
             ABORT("thread_suspend failed");
           if (GC_on_thread_event)
@@ -643,7 +642,7 @@ GC_INNER void GC_stop_world(void)
   }
 
 # ifdef MPROTECT_VDB
-    if(GC_incremental) {
+    if (GC_auto_incremental) {
       GC_mprotect_stop();
     }
 # endif
@@ -663,7 +662,11 @@ GC_INLINE void GC_thread_resume(thread_act_t thread)
   kern_return_t kern_result;
 # if defined(DEBUG_THREADS) || defined(GC_ASSERTIONS)
     struct thread_basic_info info;
-    mach_msg_type_number_t outCount = THREAD_INFO_MAX;
+    mach_msg_type_number_t outCount = THREAD_BASIC_INFO_COUNT;
+
+#   ifdef CPPCHECK
+      info.run_state = 0;
+#   endif
     kern_result = thread_info(thread, THREAD_BASIC_INFO,
                               (thread_info_t)&info, &outCount);
     if (kern_result != KERN_SUCCESS)
@@ -675,10 +678,11 @@ GC_INLINE void GC_thread_resume(thread_act_t thread)
 # endif
   /* Resume the thread */
   kern_result = thread_resume(thread);
-  if (kern_result != KERN_SUCCESS)
-    ABORT("thread_resume failed");
-  if (GC_on_thread_event)
+  if (kern_result != KERN_SUCCESS) {
+    WARN("thread_resume(%p) failed: mach port invalid\n", thread);
+  } else if (GC_on_thread_event) {
     GC_on_thread_event(GC_EVENT_THREAD_UNSUSPENDED, (void *)(word)thread);
+  }
 }
 
 /* Caller holds allocation lock, and has held it continuously since     */
@@ -690,15 +694,14 @@ GC_INNER void GC_start_world(void)
     GC_log_printf("World starting\n");
 # endif
 # ifdef MPROTECT_VDB
-    if(GC_incremental) {
+    if (GC_auto_incremental) {
       GC_mprotect_resume();
     }
 # endif
 
   if (GC_query_task_threads) {
 #   ifndef GC_NO_THREADS_DISCOVERY
-      int i;
-      int j = GC_mach_threads_count;
+      int i, j;
       kern_return_t kern_result;
       thread_act_array_t act_list;
       mach_msg_type_number_t listcount;
@@ -707,39 +710,44 @@ GC_INNER void GC_start_world(void)
       if (kern_result != KERN_SUCCESS)
         ABORT("task_threads failed");
 
-      for (i = 0; i < (int)listcount; i++) {
-        thread_act_t thread = act_list[i];
-        int last_found = j;        /* The thread index found during the   */
-                                   /* previous iteration (count value     */
-                                   /* means no thread found yet).         */
+      j = (int)listcount;
+      for (i = 0; i < GC_mach_threads_count; i++) {
+        thread_act_t thread = GC_mach_threads[i].thread;
 
-        /* Search for the thread starting from the last found one first.  */
-        while (++j < GC_mach_threads_count) {
-          if (GC_mach_threads[j].thread == thread)
-            break;
-        }
-        if (j >= GC_mach_threads_count) {
-          /* If not found, search in the rest (beginning) of the list.    */
-          for (j = 0; j < last_found; j++) {
-            if (GC_mach_threads[j].thread == thread)
+        if (GC_mach_threads[i].suspended) {
+          int last_found = j;   /* The thread index found during the    */
+                                /* previous iteration (count value      */
+                                /* means no thread found yet).          */
+
+          /* Search for the thread starting from the last found one first. */
+          while (++j < (int)listcount) {
+            if (act_list[j] == thread)
               break;
           }
-        }
-
-        if (j != last_found) {
-          /* The thread is found in GC_mach_threads.      */
-          if (GC_mach_threads[j].already_suspended) {
-#           ifdef DEBUG_THREADS
-              GC_log_printf("Not resuming already suspended thread %p\n",
-                            (void *)(word)thread);
-#           endif
-          } else {
+          if (j >= (int)listcount) {
+            /* If not found, search in the rest (beginning) of the list. */
+            for (j = 0; j < last_found; j++) {
+              if (act_list[j] == thread)
+                break;
+            }
+          }
+          if (j != last_found) {
+            /* The thread is alive, resume it.  */
             GC_thread_resume(thread);
           }
+        } else {
+          /* This thread was failed to be suspended by GC_stop_world,   */
+          /* no action needed.                                          */
+#         ifdef DEBUG_THREADS
+            GC_log_printf("Not resuming thread %p as it is not suspended\n",
+                          (void *)(word)thread);
+#         endif
         }
-
         mach_port_deallocate(my_task, thread);
       }
+
+      for (i = 0; i < (int)listcount; i++)
+        mach_port_deallocate(my_task, act_list[i]);
       vm_deallocate(my_task, (vm_address_t)act_list,
                     sizeof(thread_t) * listcount);
 #   endif /* !GC_NO_THREADS_DISCOVERY */
